@@ -1,6 +1,7 @@
+from typing import Iterable, Optional, Tuple, Union
+
 from pyproj import CRS, Transformer
 from pyproj.enums import TransformDirection
-from typing import Iterable, Optional, Tuple, Union
 
 from csrspy.enums import CoordType, VerticalDatum, Reference
 from csrspy.factories import HelmertFactory, VerticalGridShiftFactory
@@ -9,26 +10,31 @@ T_Coord3D = Tuple[float, float, float]
 T_Coord4D = Tuple[float, float, float, float]
 
 
-class ITRFtoNAD83:
+class _ToNAD83:
     direction = TransformDirection.FORWARD
 
-    def __init__(self, s_ref_frame: Union[Reference, str], s_coords: Union[str, CoordType], s_epoch: float,
+    def __init__(self, s_ref_frame: Union[Reference, str], s_coords: Union[str, CoordType],
+                 s_epoch: float, s_vd: Union[VerticalDatum, str] = VerticalDatum.GRS80,
                  t_coords: Union[str, CoordType] = None, t_epoch: float = None,
                  t_vd: Union[VerticalDatum, str] = VerticalDatum.GRS80,
                  epoch_shift_grid: str = "ca_nrc_NAD83v70VG.tif"):
         super().__init__()
-        self.s_ref_frame = s_ref_frame
+        self.s_ref_frame = Reference.ITRF14 if s_ref_frame == Reference.WGS84 else s_ref_frame
         self.s_coords = s_coords
         self.t_coords = t_coords if t_coords is not None else s_coords
         self.s_epoch = s_epoch
         self.t_epoch = t_epoch if t_epoch is not None else s_epoch
+        self.s_vd = s_vd
         self.t_vd = t_vd
         self.epoch_shift_grid = epoch_shift_grid
 
         self.transforms = []
 
-        # 1. ITRFxx Ellips -> ECEF GRS80
-        in_crs = CRS.from_proj4(self._coord_type_to_proj4(self.s_coords))
+        # 1. ITRFxx GRS80 / WGS84  -> ECEF GRS80
+        in_crs = CRS.from_proj4(self._coord_type_to_proj4(
+            self.s_coords,
+            VerticalDatum.WGS84 if self.s_vd == VerticalDatum.WGS84 else VerticalDatum.GRS80)
+        )
         grs80_crs = CRS.from_proj4("+proj=cart +ellps=GRS80")
         transform_in2cartestian = Transformer.from_crs(in_crs, grs80_crs)
         self.transforms.append(transform_in2cartestian)
@@ -57,14 +63,17 @@ class ITRFtoNAD83:
         self.transforms.append(transform_out)
 
     @staticmethod
-    def _coord_type_to_proj4(coord_type: CoordType) -> str:
+    def _coord_type_to_proj4(coord_type: Union[CoordType, str],
+                             ellps: Union[VerticalDatum, str] = VerticalDatum.GRS80) -> str:
+        ellps = ellps.upper()
+
         if coord_type == CoordType.GEOG:
-            return "+proj=longlat +datum=WGS84 +no_defs"
+            return f"+proj=longlat +ellps={ellps} +no_defs"
         if coord_type == CoordType.CART:
-            return "+proj=cart +datum=WGS84 +no_defs"
+            return f"+proj=cart +ellps={ellps} +no_defs"
         else:
             zone = int(coord_type.value[3:])
-            return f"+proj=utm +zone={zone} +datum=WGS84 +units=m +no_defs"
+            return f"+proj=utm +zone={zone} +ellps={ellps} +units=m +no_defs"
 
     def _coord_3d_to_4d(self, coord: T_Coord3D) -> T_Coord4D:
         return coord[0], coord[1], coord[2], self.s_epoch
@@ -85,14 +94,18 @@ class ITRFtoNAD83:
         return map(self._coord_4d_to_3d, coords)
 
 
-class NAD83toITRF(ITRFtoNAD83):
-    """This class is the same as ITRFtoNAD83, but does all transformations in reverse."""
+class _FromNAD83(_ToNAD83):
+    """This class is the same as toNAD83, but does all transformations in reverse."""
     direction = TransformDirection.INVERSE
 
-    def __init__(self, t_ref_frame: Union[Reference, str], t_coords: Union[str, CoordType], t_epoch: float,
+    def __init__(self, t_ref_frame: Union[Reference, str], t_coords: Union[str, CoordType],
+                 t_epoch: float, t_vd: Union[VerticalDatum, str] = VerticalDatum.GRS80,
                  s_coords: Union[str, CoordType] = None, s_epoch: float = None,
-                 s_vd: Union[VerticalDatum, str] = VerticalDatum.GRS80, epoch_shift_grid: str = "ca_nrc_NAD83v70VG.tif"):
-        super().__init__(t_ref_frame, t_coords, t_epoch, s_coords, s_epoch, s_vd, epoch_shift_grid)
+                 s_vd: Union[VerticalDatum, str] = VerticalDatum.GRS80,
+                 epoch_shift_grid: str = "ca_nrc_NAD83v70VG.tif"):
+        super().__init__(s_ref_frame=t_ref_frame, s_coords=t_coords, s_epoch=t_epoch,
+                         s_vd=t_vd, t_coords=s_coords, t_epoch=s_epoch, t_vd=s_vd,
+                         epoch_shift_grid=epoch_shift_grid)
         self.transforms.reverse()
 
 
@@ -129,34 +142,42 @@ class CSRSTransformer(object):
         self.validate_crs(s_ref_frame, s_vd)
         self.validate_crs(t_ref_frame, t_vd)
 
-        if self.is_itrf(s_ref_frame) and self.is_nad83(t_ref_frame):
-            self.transformers = [ITRFtoNAD83(s_ref_frame=self.s_ref_frame, s_coords=self.s_coords, s_epoch=self.s_epoch,
-                                             t_coords=self.t_coords, t_epoch=self.t_epoch, t_vd=self.t_vd,
-                                             epoch_shift_grid=self.epoch_shift_grid)]
+        if (not self.is_nad83(s_ref_frame)) and self.is_nad83(t_ref_frame):
+            self.transformers = [_ToNAD83(s_ref_frame=self.s_ref_frame, s_coords=self.s_coords,
+                                          s_epoch=self.s_epoch, s_vd=self.s_vd,
+                                          t_coords=self.t_coords, t_epoch=self.t_epoch,
+                                          t_vd=self.t_vd,
+                                          epoch_shift_grid=self.epoch_shift_grid)]
 
-        elif self.is_nad83(s_ref_frame) and self.is_itrf(t_ref_frame):
-            self.transformers = [NAD83toITRF(t_ref_frame=self.t_ref_frame, t_coords=self.t_coords, t_epoch=self.t_epoch,
-                                             s_coords=self.s_coords, s_epoch=self.s_epoch, s_vd=self.s_vd,
-                                             epoch_shift_grid=self.epoch_shift_grid)]
+        elif self.is_nad83(s_ref_frame) and (not self.is_nad83(t_ref_frame)):
+            self.transformers = [_FromNAD83(t_ref_frame=self.t_ref_frame, t_coords=self.t_coords,
+                                            t_epoch=self.t_epoch, t_vd=self.t_vd,
+                                            s_coords=self.s_coords, s_epoch=self.s_epoch,
+                                            s_vd=self.s_vd,
+                                            epoch_shift_grid=self.epoch_shift_grid)]
 
-        elif self.is_itrf(s_ref_frame) and self.is_itrf(t_ref_frame):
+        elif not (self.is_nad83(s_ref_frame) or self.is_nad83(t_ref_frame)):
             self.transformers = [
-                ITRFtoNAD83(s_ref_frame=self.s_ref_frame, s_coords=self.s_coords, s_epoch=self.s_epoch,
-                            t_coords=self.t_coords, t_epoch=self.t_epoch, t_vd=VerticalDatum.GRS80,
-                            epoch_shift_grid=self.epoch_shift_grid),
-                NAD83toITRF(t_ref_frame=self.t_ref_frame, t_coords=self.t_coords, t_epoch=self.t_epoch,
-                            s_coords=self.t_coords, s_epoch=self.t_epoch, s_vd=VerticalDatum.GRS80,
-                            epoch_shift_grid=self.epoch_shift_grid)
+                _ToNAD83(s_ref_frame=self.s_ref_frame, s_coords=self.s_coords,
+                         s_epoch=self.s_epoch, s_vd=self.s_vd,
+                         t_coords=self.t_coords, t_epoch=self.t_epoch, t_vd=VerticalDatum.GRS80,
+                         epoch_shift_grid=self.epoch_shift_grid),
+                _FromNAD83(t_ref_frame=self.t_ref_frame, t_coords=self.t_coords,
+                           t_epoch=self.t_epoch, t_vd=self.t_vd,
+                           s_coords=self.t_coords, s_epoch=self.t_epoch, s_vd=VerticalDatum.GRS80,
+                           epoch_shift_grid=self.epoch_shift_grid)
             ]
 
         elif self.is_nad83(s_ref_frame) and self.is_nad83(t_ref_frame):
             self.transformers = [
-                NAD83toITRF(t_ref_frame=Reference.ITRF14, t_coords=self.t_coords, t_epoch=self.t_epoch,
-                            s_coords=self.s_coords, s_epoch=self.s_epoch, s_vd=self.s_vd,
-                            epoch_shift_grid=self.epoch_shift_grid),
-                ITRFtoNAD83(s_ref_frame=Reference.ITRF14, s_coords=self.t_coords, s_epoch=self.t_epoch,
-                            t_coords=self.t_coords, t_epoch=self.t_epoch, t_vd=self.t_vd,
-                            epoch_shift_grid=self.epoch_shift_grid),
+                _FromNAD83(t_ref_frame=Reference.ITRF14, t_coords=self.t_coords,
+                           t_epoch=self.t_epoch, t_vd=self.t_vd,
+                           s_coords=self.s_coords, s_epoch=self.s_epoch, s_vd=self.s_vd,
+                           epoch_shift_grid=self.epoch_shift_grid),
+                _ToNAD83(s_ref_frame=Reference.ITRF14, s_coords=self.t_coords,
+                         s_epoch=self.t_epoch, s_vd=self.t_vd,
+                         t_coords=self.t_coords, t_epoch=self.t_epoch, t_vd=self.t_vd,
+                         epoch_shift_grid=self.epoch_shift_grid),
             ]
 
     @staticmethod
@@ -165,11 +186,22 @@ class CSRSTransformer(object):
 
     @staticmethod
     def is_itrf(ref_frame: Reference):
-        return ref_frame != Reference.NAD83CSRS
+        return ref_frame != Reference.NAD83CSRS and ref_frame != Reference.WGS84
+
+    @staticmethod
+    def is_wgs84(ref_frame: Reference):
+        return ref_frame == Reference.WGS84
 
     def validate_crs(self, ref_frame: Reference, vd: Optional[VerticalDatum]):
         if self.is_itrf(ref_frame):
-            assert vd is VerticalDatum.WGS84, f"{ref_frame} must use VerticalDatum WGS84."
+            assert vd is VerticalDatum.GRS80, \
+                f"{ref_frame} must use VerticalDatum GRS80."
+        elif self.is_wgs84(ref_frame):
+            assert (vd is VerticalDatum.WGS84) or (vd is VerticalDatum.GRS80), \
+                f"{ref_frame} must use VerticalDatum WGS84 or GRS80."
+        elif self.is_nad83(ref_frame):
+            assert vd is not VerticalDatum.WGS84, \
+                f"{ref_frame} must not use VerticalDatum WGS84."
 
     def __call__(self, coords: Iterable[T_Coord3D]) -> Iterable[T_Coord3D]:
         """
